/**********************************************/
/* PalmRControl 1.0b1 by Rod                  */
/*                                            */
/* Copyright(C) 2006 Rodolfo Turco            */
/*                                            */
/* It's under the GNU General Public License  */
/* If you derive your own project remember to */
/* change the name and the AppCreator ID      */
/*                                            */
/* http://www.webalice.it/rodolfo.turco       */
/* rodolfo.turco[ a T ] virgilio [dot] it     */
/*                                            */
/**********************************************/

//TODO 1.0
// * when opened, the state of Main isn't setted correctly

#include "PalmOS.h"
#include "common\system\palmOneNavigator.h"
#include "BtCommVdrv.h"

// ==============================================================
// This is to manage all the 8 direction in the 5-way nav,
// you may need to modify it to adapt it to your hardware
#define navBitUpRight			  (0x0001 | 0x0008)
#define navBitUpLeft			  (0x0001 | 0x0004)
#define navBitDownRight		  (0x0002 | 0x0008)
#define navBitDownLeft		  (0x0002 | 0x0004)
#define navChangeUpRight    (0x0100 | 0x0800)
#define navChangeUpLeft	    (0x0100 | 0x0400)
#define navChangeDownRight  (0x0200 | 0x0800)
#define navChangeDownLeft	  (0x0200 | 0x0400)
#define NavDirectionPressedEx(eventP, nav)                                               \
(                                                                                        \
    IsFiveWayNavEvent(eventP)                                                            \
        ? (   NavDirectionPalmPressed(eventP, nav) )                                     \
        : (   ((eventP)->data.keyDown.chr == vchrPageUp && (navBit ## nav) == navBitUp)  \
	 	   || ((eventP)->data.keyDown.chr == vchrPageDown && (navBit ## nav) == navBitDown)) \
)
#define NavKeyPressedEx(eventP, nav)                                                     \
(                                                                                        \
    (navBit ## nav == navBitSelect)                                                      \
        ? NavSelectPressed(eventP)                                                       \
        : NavDirectionPressedEx(eventP, nav)                                             \
)

// ==============================================================
// Forms defines
#define MainForm         1000
#define CMDHLevel        1001
#define CMDPrograde      1002
#define CMDRetrograde    1003
#define CMDNormalPlus    1004
#define CMDNormalMinus   1005

#define BMPLEDOFF        1000
#define BMPLEDLOCALON    1100
#define BMPLEDON         1200
#define BMPLEDRECEIVING  1300

#define ScrollRCSLevel   1007
#define CMDLinF          1008
#define CMDLinB          1009
#define CMDLinD          1010
#define CMDLinR          1011
#define CMDLinU          1012
#define CMDLinL          1013
#define CMDRCSLin        1014
#define CMDRotX          1015
#define CMDRotRr         1016
#define CMDRotD          1017
#define CMDRotR          1018
#define CMDRotU          1019
#define CMDRotL          1020
#define CMDShutdown      1021
#define CMDRotRl         1022
#define ScrollHover      1023
#define ScrollRetro      1024
#define ScrollMain       1025
#define CMDSetFocus      1027
#define CMDRCSRot        1029
#define VesselList       1030
#define VesselPopTrig    1031
#define CMDKillRot       1041
#define CMDLinX          1042
#define CMDHAlt          1043
#define CMDRCSOff        1044
#define RomIncompAlert   1100
#define AlertGenericErr  1101

#define MenuOpenCOM      1300
#define MenuHelp         1301
#define MenuExit         1302
#define MenuAbout        1303
#define MenuCloseCOM     1304
#define MenuPreferences  1305

#define PrefsForm        1500
#define PrefsOk          1501
#define PrefsClear       1502
#define PrefsBaudRate    1503
#define PrefsPort        1504
#define PrefsKeepAlive   1508
#define Prefs5Way        1510
#define PrefsSimple      1511
#define PrefsInvDirect   1512

#define AboutForm        1600
#define AboutFormOK      1601
#define HELP_TEXT        1900

#define AppCreator          'PRCo'
#define sysFileTApplication 'appl'
#define MinVersion		      0x03500000

#define BUFFERSIZE      4096 // Commands Buffer
#define RXBUFFERSIZE     512
#define MAXNUMVESSELS     32
#define MAXSIZEVESSELNAME 20
#define COMMANDTERMINATOR   (unsigned char)23 // is the command terminator (ASCII CODE 23)
#define STRINGTERMINATOR    (unsigned char)34 // is the string terminator (ASCII CODE 34)

// ==============================================================
// COM commands
// ==============================================================

// To Palm
#define COMMAND_SV 0x5653 // Set Vessel list

// From/To Palm
#define COMMAND_CO 0x434F // Set COM Open (advide the other side it's ready to communicate)
#define COMMAND_CL 0x434C // Set COM Close (advide the other side it closed the connection)
#define COMMAND_Ho 0x486F // Set Hover
#define COMMAND_Ma 0x4D61 // Set Main
#define COMMAND_Rt 0x5274 // Set Retro
#define COMMAND_KR 0x4B52 // Set KillRot
#define COMMAND_HL 0x484C // Set HLevel
#define COMMAND_HA 0x4841 // Set HAlt
#define COMMAND_Pr 0x5072 // Set Prograde
#define COMMAND_Re 0x5265 // Set Retrograde
#define COMMAND_NP 0x4E50 // Set Normal +
#define COMMAND_NM 0x4E4D // Set Normal -
#define COMMAND_Le 0x4C65 // Set RCS Level
#define COMMAND_SR 0x5352 // Set RCS Mode (0=disable, 1=rotational, 2=linear)
#define COMMAND_Se 0x5365 // Select vessel

// From Palm
#define COMMAND_RD 0x5244 // Set RCS Rot D
#define COMMAND_RU 0x5255 // Set RCS Rot U
#define COMMAND_Rl 0x526C // Set RCS Rot Left ^
#define COMMAND_Rr 0x5272 // Set RCS Rot Right ^
#define COMMAND_RL 0x524C // Set RCS Rot Left
#define COMMAND_RR 0x5252 // Set RCS Rot Right
#define COMMAND_RX 0x5258 // Set RCS Rot X
#define COMMAND_LD 0x4C44 // Set RCS Lin D
#define COMMAND_LU 0x4C55 // Set RCS Lin U
#define COMMAND_LF 0x4C46 // Set RCS Lin F
#define COMMAND_LB 0x4C42 // Set RCS Lin B
#define COMMAND_LL 0x4C4C // Set RCS Lin L
#define COMMAND_LR 0x4C52 // Set RCS Lin R
#define COMMAND_SF 0x5346 // Set Focus to vessel
#define COMMAND_GV 0x4756 // Get Vessel list
#define COMMAND_GH 0x4748 // Get Hover
#define COMMAND_GM 0x474D // Get Main
#define COMMAND_GR 0x4752 // Get Retro
#define COMMAND_GL 0x474C // Get RCS Level
#define COMMAND_Gi 0x4769 // Get KillRot
#define COMMAND_Ge 0x4765 // Get HLevel
#define COMMAND_Gh 0x4768 // Get HAlt
#define COMMAND_Go 0x476F // Get Prograde
#define COMMAND_Gt 0x4774 // Get Retrograde
#define COMMAND_Gp 0x4770 // Get Normal +
#define COMMAND_Gm 0x476D // Get Normal -
#define COMMAND_GT 0x4754 // Get RCS type
#define COMMAND_GF 0x4746 // Get Focused vessel


/************************************/
/* Functions protos                 */
/************************************/
static Err RomVersionCompatible(UInt32 requiredVersion,UInt16 launchFlags);
static void *getObjectPtr(FormPtr frmP,UInt16 unID);
void ClampValue16(UInt16 *punValue,UInt16 Min,UInt16 Max);
void SetValue16(UInt16 unID,UInt16 unValue);
UInt16 GetValue16(UInt16 unID,UInt16 Min,UInt16 Max);
void GetString(UInt16 unID,char *cString,UInt16 unLen);
void SetString(UInt16 unID,char *cString);
UInt32 TimeGetMilliSeconds();
#define SBLEFTRIGHT 0 // ScrollBarHighlight direction left right
#define SBRIGHTLEFT 1 // ScrollBarHighlight direction right left
#define SBDOWNUP    2 // ScrollBarHighlight direction down up
void ScrollBarHighlight(UInt16 unTop,UInt16 unLeft,UInt16 unWidth,UInt32 unValue,UInt32 unMaxValue,UInt16 unDir);
static Boolean doMenu(FormPtr frmP, UInt16 command);
Boolean PreprocessEvent(EventPtr eventP);
Boolean mainFormEventHandler(EventPtr eventP);
void startApp();
void stopApp();
Boolean appHandleEvent(EventPtr event);
UInt32 PilotMain(UInt16 cmd,void *cmdPBP,UInt16 launchFlags);
void	setList(Char **itemsText,Int16 numItems,UInt16 unID,UInt16 unDim);
void updateVesselList();
void drawVesselList();
void setFocusVessel();
void GetListItem(UInt16 unListID, char * cStr);
void COMSendString(short key, char* pStringa);
void emptyVesselList();
void COMSendSimplex(short key);

// ==============================================================
// Application preferences

// Prefs.ulControlBits:
#define CBITS_KEEPALIVE  0x08000000 // Keep alive
#define CBITS_SIMPLEKEYB 0x04000000 // Simple keyb
#define CBITS_INVDIRECT  0x02000000 // Inverted directions

typedef struct {
 UInt32 ulSerialPort;
 UInt32 ulBaudRate;
 UInt32 ulControlBits; //See above
 UInt16 unKeybMode;
} Prefs;

/************************************/
/* Global variables                 */
/************************************/
static Int16   nRCSMode;
static UInt16  unTickPerSec;
static UInt16  unPortId;
static UInt16  initDelayDef;
static UInt16  periodDef;
static UInt16  doubleTapDelayDef;
static UInt16  initDelay;
static UInt16  period;
static UInt16  doubleTapDelay;
static UInt16  unBufferIx;
static UInt16  unVesselListItem;
static UInt16  unIx;
static UInt32  ulTime;
static UInt32  ulVesselListNum;
static UInt32  ulOldTime;
static Boolean bCanDraw;
static Boolean bReceivingLed;
static Boolean bRemoteOpen;
static Boolean bKeepAlive;
static Boolean bSimpleKeyb;
static Boolean bInvDirect;
static Boolean bNewSimpleKeyb;
static Boolean queueAheadDef;
static Boolean queueAhead;
static Boolean bCMDHLevel;
static Boolean bCMDKillRot;
static Boolean bCMDPrograde;
static Boolean bCMDRetrograde;
static Boolean bCMDNormalPlus;
static Boolean bCMDNormalMinus;
static Boolean bCMDHAlt;
static Boolean bKeyRatesUp;
static Char    cSpaces[80];
static Char    *ppVesselList[MAXNUMVESSELS];
static Char    cVesselList[MAXNUMVESSELS][MAXSIZEVESSELNAME];
static Int8    cFocusVessel[MAXSIZEVESSELNAME];
static Int8    cStringa[256];
static UInt8*  cCommandBuffer;
static UInt8   cRxBuffer[RXBUFFERSIZE];
static Prefs   prefs;
static Err     Errore;
static MemHandle     hBuffer;
static RectangleType Rect;
static RGBColorType  newRgb;
static RGBColorType  prevRgb;
static MemHandle     hRsc;
static BitmapType*   bitmapP;

//#define _PRCoDEBUG

#ifdef _PRCoDEBUG
static FileHand  hFile;
#endif

/************************************/
/* Checks the Rom version           */
/************************************/
static Err RomVersionCompatible(UInt32 requiredVersion,UInt16 launchFlags)
{
	UInt32 romVersion;

	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if(romVersion < requiredVersion)
		{
		if((launchFlags&(sysAppLaunchFlagNewGlobals|sysAppLaunchFlagUIApp))==
			(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
			{
    FrmAlert(RomIncompAlert);
			 if (romVersion < 0x02000000)
				 {				
				  AppLaunchWithCommand(sysFileCDefaultApp,sysAppLaunchCmdNormalLaunch,NULL);
				 }
			}
		return (sysErrRomIncompatible);
		}
	return 0;
}

/**************************************************/
/* Get the Object pointer needed to some function */
/**************************************************/
static void *getObjectPtr(FormPtr frmP,UInt16 unID)
{
 UInt16 fldIndex=FrmGetObjectIndex(frmP,unID);
 return FrmGetObjectPtr(frmP,fldIndex);
}

/************************************/
static void DrawLed(UInt16 unBitmapID)
{
 if(!bCanDraw)
  return;

 hRsc = DmGetResource(bitmapRsc, unBitmapID);
 if(hRsc)
 {
  bitmapP = (BitmapType*) MemHandleLock(hRsc);
  if(bitmapP)
   WinDrawBitmap(bitmapP, 148, 0);
  MemHandleUnlock(hRsc);
  DmReleaseResource(hRsc);
 }
}

/************************************/
/* Clamp value in a range           */
/************************************/
void ClampValue16(UInt16 *punValue,UInt16 Min,UInt16 Max)
 {
  if(*punValue<Min)
   *punValue=Min;
  else if(*punValue>Max)
   *punValue=Max;
 }

/************************************/
/* Set a numeric value in an edit   */
/************************************/
void SetValue16(UInt16 unID,UInt16 unValue)
 {
  MemHandle h, oldHandle;
  char *p;
  FieldType* fldP;
  FormType* frmP;
  char cStrTemp[8];
 
  frmP = FrmGetActiveForm();
  h = MemHandleNew(8);
  p = (char *)MemHandleLock(h);
  StrIToA(cStrTemp,unValue);
  StrCopy(p,cStrTemp);
  MemHandleResize(h,StrLen(p)+1);
  MemHandleUnlock(h);
  fldP=(FieldType*)FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,unID));
  oldHandle = FldGetTextHandle(fldP);
  FldSetTextHandle(fldP, h);
  if(oldHandle)
   MemHandleFree(oldHandle);
  FldEraseField(fldP);
  FldDrawField(fldP);
 }

/************************************/
/* Get a number from an edit        */
/************************************/
UInt16 GetValue16(UInt16 unID,UInt16 Min,UInt16 Max)
{
 UInt16 unValue=0;
 FormPtr frmP=FrmGetActiveForm();
 FieldPtr fldP=getObjectPtr(frmP,unID);
 MemHandle h=FldGetTextHandle(fldP);
 if(h!=NULL)
  {
   char *p = MemHandleLock(h);
   unValue=StrAToI(p);
   MemHandleUnlock(h);
   ClampValue16(&unValue,Min,Max);
  }
 return unValue;
}

/************************************/
/* Set a numeric value in an edit   */
/************************************/
void SetValue32(UInt16 unID, UInt32 ulValue)
{
  MemHandle h, oldHandle;
  char *p;
  FieldType* fldP;
  FormType* frmP;
  char cStrTemp[12];

  frmP = FrmGetActiveForm();
  h = MemHandleNew(12);
  p = (char *)MemHandleLock(h);
  StrIToA(cStrTemp,ulValue);
  StrCopy(p,cStrTemp);
  MemHandleResize(h,StrLen(p)+1);
  MemHandleUnlock(h);
  fldP=(FieldType*)FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,unID));
  oldHandle = FldGetTextHandle(fldP);
  FldSetTextHandle(fldP, h);
  if(oldHandle)
    MemHandleFree(oldHandle);
  FldEraseField(fldP);
  FldDrawField(fldP);
}
 
/************************************/
/* Get a number from an edit        */
/************************************/
UInt32 GetValue32(UInt16 unID)
{
  UInt32 ulValue=0;

  FormPtr frmP=FrmGetActiveForm();
  FieldPtr fldP=getObjectPtr(frmP,unID);
  MemHandle h=FldGetTextHandle(fldP);
  if(h!=NULL)
  {
    char *p = MemHandleLock(h);
    ulValue=StrAToI(p);
    MemHandleUnlock(h);
  }
  return ulValue;
}

/**********************************/
/* Set a string on an edit field  */
/**********************************/
void SetString(UInt16 unID,char *cString)
{
 char *p;
 UInt16 unLen;
 MemHandle h;
 FormPtr frmP = FrmGetActiveForm();
 FieldPtr fldP = getObjectPtr(frmP,unID);
 MemHandle oldH = FldGetTextHandle(fldP);

 unLen=StrLen(cString);
 h = MemHandleNew(unLen + 1);
 p = (char *)MemHandleLock(h);
 StrCopy(p,cString);
 MemHandleUnlock(h);
 FldSetTextHandle(fldP, h);
 FldDrawField(fldP);
 if(oldH)
  MemHandleFree(oldH);
}

/*************************************************************************/
/* Get the time in milliseconds with a tick resolution (tipically 10 ms) */
/*************************************************************************/
UInt32 TimeGetMilliSeconds()
 {
  if(unTickPerSec)
   return(TimGetTicks()*1000/unTickPerSec);
  else
   return(TimGetSeconds()*1000);
 }

/*******************************************************/
void ScrollBarHighlight(UInt16 unTop,UInt16 unLeft,UInt16 unWidth,UInt32 unValue,UInt32 unMaxValue,UInt16 unDir)
 {
  UInt32 ulTemp=0;

  if(!bCanDraw)
  return;

  if(unMaxValue)
   ulTemp=unWidth*unValue/unMaxValue;
  if(ulTemp>unWidth)
   ulTemp=unWidth;
  switch(unDir)
  {
    case SBLEFTRIGHT:
      Rect.topLeft.x=unLeft;
      Rect.topLeft.y=unTop;
      Rect.extent.x=ulTemp;
      Rect.extent.y=6;
      break;
    case SBRIGHTLEFT:
      Rect.topLeft.x=unLeft-ulTemp;
      Rect.topLeft.y=unTop;
      Rect.extent.x=ulTemp;
      Rect.extent.y=6;
      break;
    case SBDOWNUP:
      Rect.topLeft.x=unLeft;
      Rect.topLeft.y=unTop-ulTemp;
      Rect.extent.x=6;
      Rect.extent.y=ulTemp;
      break;
  }
  newRgb.r=0xFF;
  newRgb.g=0x00;
  newRgb.b=0x00;
  WinSetForeColorRGB(&newRgb,&prevRgb);
  WinDrawRectangle(&Rect,0); 
 }

/*************************************************/
static void CloseSerial()
{
  if(unPortId)
  {
    COMSendSimplex(COMMAND_CL);
    DrawLed(BMPLEDOFF);
    if(bCanDraw)
      WinDrawChars("< COM Inactive >",16,19,2);
    KeyRates(true, &initDelayDef, &periodDef, &doubleTapDelayDef, &queueAheadDef);
    bKeyRatesUp = false;
    emptyVesselList();
    SysTaskDelay(unTickPerSec/2); // 500 ms
    SrmClose(unPortId);
    unPortId = 0;
  }

#ifdef _PRCoDEBUG
if(hFile)
  {
    FileClose(hFile);
    hFile = 0;
  }
#endif
}

/************************************/
void COMSendSimplex(short key)
{
  if(unPortId)
  {
    *(UInt16 *)cStringa = key;
    cStringa[2] = COMMANDTERMINATOR;
    cStringa[3] = 0;
    SrmSend(unPortId, cStringa, 3, &Errore);
#ifdef _PRCoDEBUG
 if(hFile)
 {
  char cTemp[2];
  cTemp[0] = '[';
  FileWrite(hFile,cTemp,1,1,&Errore);
  FileWrite(hFile,cStringa,3,1,&Errore);
  cTemp[0] = ']';
  FileWrite(hFile,cTemp,1,1,&Errore);
 }
#endif
  }
}

/************************************/
void COMSendNumber(short key, long lNumber)
{
  char  cStrTemp[16];
  Int16 nLen;

  if(unPortId)
  {
    *(UInt16 *)cStringa = key;
    cStringa[2] = 0;
    StrIToA(cStrTemp,lNumber);
    StrCat(cStringa, cStrTemp);
    nLen = StrLen(cStringa);
    cStringa[nLen] = COMMANDTERMINATOR;
    cStringa[nLen+1] = 0;
    SrmSend(unPortId, cStringa, nLen+1, &Errore);
#ifdef _PRCoDEBUG
 if(hFile)
 {
  char cTemp[2];
  cTemp[0] = '[';
  FileWrite(hFile,cTemp,1,1,&Errore);
  FileWrite(hFile,cStringa,nLen+1,1,&Errore);
  cTemp[0] = ']';
  FileWrite(hFile,cTemp,1,1,&Errore);
 }
#endif
  }
}

/************************************/
void COMSendString(short key, char* pStringa)
{
  Int16 nLen;
  unsigned char cStringa[256];

  if(unPortId)
  {
    *(UInt16 *)cStringa = key;
    cStringa[2] = 0;
    StrCat(cStringa, pStringa);
    nLen = StrLen(cStringa);
    cStringa[nLen] = COMMANDTERMINATOR;
    cStringa[nLen+1] = 0;
    SrmSend(unPortId, cStringa, nLen+1, &Errore);
#ifdef _PRCoDEBUG
 if(hFile)
 {
  char cTemp[2];
  cTemp[0] = '[';
  FileWrite(hFile,cTemp,1,1,&Errore);
  FileWrite(hFile,cStringa,nLen+1,1,&Errore);
  cTemp[0] = ']';
  FileWrite(hFile,cTemp,1,1,&Errore);
 }
#endif
  }
}

/*************************************************/
static void OpenSerial()
{
  CloseSerial();

#ifdef _PRCoDEBUG
 hFile = FileOpen(0,"PRCtrl",'Log ','LgrP',fileModeReadWrite,&Errore);
 if(Errore == errNone)
  FileSeek(hFile,0,fileOriginBeginning);
#endif

  unBufferIx = 0;
  cCommandBuffer[0] = 0;
  Errore = SrmOpen(prefs.ulSerialPort, prefs.ulBaudRate, &unPortId);
  if(Errore == errNone)
  {
    SrmReceiveFlush(unPortId, 300);
    if(bCanDraw)
     WinDrawChars(cSpaces,44,19,2);
    COMSendSimplex(COMMAND_CO);//Ready to COM
    initDelay = 1;
    period = 1;
    doubleTapDelay = 3;
    queueAhead = false;
    KeyRates(true, &initDelay, &period, &doubleTapDelay, &queueAhead);
    bKeyRatesUp = true;
    DrawLed(BMPLEDLOCALON);
  }
  else
  {
    DrawLed(BMPLEDOFF);
    unPortId = 0;
    FrmCustomAlert(AlertGenericErr,"Can't open serial",NULL,NULL);
  }
}

/************************************/
/* Menu events management           */
/************************************/
static Boolean doMenu(FormPtr frmP, UInt16 command)
{
 EventType event;
 Boolean handled = false;

 switch(command)
  {
   case MenuAbout:
    FrmGotoForm(AboutForm);
    handled = true; 
    break;
   case MenuPreferences:
    FrmGotoForm(PrefsForm);
    handled = true; 
    break;
   case MenuHelp:
    if(bKeyRatesUp)
      KeyRates(true, &initDelayDef, &periodDef, &doubleTapDelayDef, &queueAheadDef);
    FrmHelp(HELP_TEXT);
    if(bKeyRatesUp)
    {
      initDelay = 1;
      period = 1;
      doubleTapDelay = 3;
      queueAhead = false;
      KeyRates(true, &initDelay, &period, &doubleTapDelay, &queueAhead);
    }
    handled = true; 
    break;
   case MenuOpenCOM:
    OpenSerial();
    handled = true; 
    break;
   case MenuCloseCOM:
    CloseSerial();
    handled = true; 
    break;
   case MenuExit:
    MemSet(&event,sizeof(event),0);
    event.eType = keyDownEvent;
    event.data.keyDown.chr = vchrLaunch;
    event.data.keyDown.modifiers = commandKeyMask;
    EvtAddEventToQueue(&event);
    handled = true;
    break;
   default:
    break;
  }
 return handled;
}

/*************************************************/
void changeRCSmode()
{
  if(!bCanDraw)
    return;

  switch(nRCSMode)
  {
  case 2: // -> 1 Rot
    nRCSMode = 1;
    WinDrawInvertedChars("Rot",3,59,49);
    WinDrawChars("Lin",3,124,49);
    break;
  case 1: // -> 2 Lin
    nRCSMode = 2;
    WinDrawChars("Rot",3,59,49);
    WinDrawInvertedChars("Lin",3,124,49);
    break;
  case 0: // -> 0 None
    nRCSMode = 0;
    WinDrawChars("Rot",3,59,49);
    WinDrawChars("Lin",3,124,49);
    break;
  }
  COMSendNumber(COMMAND_SR, nRCSMode);
}

/*************************************************/
void setFocusVessel()
{
  for(unIx=0; unIx<ulVesselListNum; unIx++)
  {
    if(StrCompareAscii(cFocusVessel, cVesselList[unIx]) == 0)
    {
      unVesselListItem = unIx;
      break;
    }
  }
}

/*************************************************/
void emptyVesselList()
{
  for(unVesselListItem=0; unVesselListItem<MAXNUMVESSELS; unVesselListItem++)
  {
   cVesselList[unVesselListItem][0]=0;
  }
  ulVesselListNum = 0;
  unVesselListItem = 0;
  cFocusVessel[0] = 0;
}

/*************************************************/
/* You may need to preprocess events, do it here */
/*************************************************/
Boolean PreprocessEvent(EventPtr eventP)
{
 Boolean handled = false;
 Boolean bReinterpret;
 Boolean bCommandTerminator;
 WChar   chrPress;
 Int16   nSel;
 UInt16  unTempIx;
 UInt16  unDestIx;
 UInt16  unCommand;
 Int32   lTotReadByte;
 ScrollBarPtr pScl;
 FormPtr frmP = FrmGetActiveForm();

 nSel = 0;
 chrPress=eventP->data.keyDown.chr;

 switch(eventP->eType)
  {
   case winExitEvent:
    if(eventP->data.winExit.exitWindow==(WinHandle)FrmGetFormPtr(MainForm))
     bCanDraw=false;
    break;
   case winEnterEvent:
    if(eventP->data.winEnter.enterWindow==(WinHandle)FrmGetFirstForm())
     {
      if(eventP->data.winEnter.enterWindow==(WinHandle)FrmGetFormPtr(MainForm))
       bCanDraw=true;
     }
    break;
   case keyDownEvent:
    if(EvtKeydownIsVirtual(eventP))
       {
        if(NavKeyPressed(eventP,Select))
         {
          bCMDKillRot = true;
          if(bCanDraw)
            WinDrawInvertedChars("KillRot",7,16,131);
          COMSendSimplex(COMMAND_RX);
          COMSendSimplex(COMMAND_Ge); // Get HLevel
          COMSendSimplex(COMMAND_Gh); // Get HAlt
          handled=true;
          break;
         }
        if(NavKeyPressed(eventP, Left))
         {
           if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_Rl);
           else if(nRCSMode == 2) // Lin
             COMSendSimplex(COMMAND_LL);
          handled=true;
          break;
         }
        if(NavKeyPressed(eventP, Right))
         {
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_Rr);
           else if(nRCSMode == 2) // Lin
             COMSendSimplex(COMMAND_LR);
          handled=true;
          break;
         }
        if(NavKeyPressed(eventP, Up))
         {
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_RD);
           else if(nRCSMode == 2) // Lin
           {
             if(bInvDirect)
              COMSendSimplex(COMMAND_LU);
             else
              COMSendSimplex(COMMAND_LD);
          }
          handled=true;
          break;
         }
        if(NavKeyPressed(eventP, Down))
         {
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_RU);
           else if(nRCSMode == 2) // Lin
           {
             if(bInvDirect)
              COMSendSimplex(COMMAND_LD);
             else
              COMSendSimplex(COMMAND_LU);
           }
          handled=true;
          break;
         }
        if(NavKeyPressedEx(eventP, UpLeft))
        {
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_RD);
           else if(nRCSMode == 2) // Lin
           {
             if(bInvDirect)
              COMSendSimplex(COMMAND_LU);
             else
              COMSendSimplex(COMMAND_LD);
           }
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_Rl);
           else if(nRCSMode == 2) // Lin
             COMSendSimplex(COMMAND_LL);
          handled=true;
          break;
        }
       if(NavKeyPressedEx(eventP, UpRight))
        {
          if(nRCSMode == 1) // Rot
            COMSendSimplex(COMMAND_RD);
          else if(nRCSMode == 2) // Lin
           {
             if(bInvDirect)
              COMSendSimplex(COMMAND_LU);
             else
              COMSendSimplex(COMMAND_LD);
           }
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_Rr);
           else if(nRCSMode == 2) // Lin
             COMSendSimplex(COMMAND_LR);
          handled=true;
          break;
        }
       if(NavKeyPressedEx(eventP, DownLeft))
        {
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_RU);
           else if(nRCSMode == 2) // Lin
           {
             if(bInvDirect)
              COMSendSimplex(COMMAND_LD);
             else
              COMSendSimplex(COMMAND_LU);
           }
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_Rl);
           else if(nRCSMode == 2) // Lin
             COMSendSimplex(COMMAND_LL);
          handled=true;
          break;
        }
       if(NavKeyPressedEx(eventP, DownRight))
        {
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_RU);
           else if(nRCSMode == 2) // Lin
           {
             if(bInvDirect)
              COMSendSimplex(COMMAND_LD);
             else
              COMSendSimplex(COMMAND_LU);
          }
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_Rr);
           else if(nRCSMode == 2) // Lin
             COMSendSimplex(COMMAND_LR);
          handled=true;
          break;
        }
       }
    if((((eventP)->data.keyDown.modifiers & autoRepeatKeyMask) != 0) &&
       (((eventP)->data.keyDown.keyCode & navBitsAll) == navBitSelect))
      {
        // Select button long press, we don't want to return to the launcher
        handled=true;
        break;
      }
    switch(chrPress)
     {
      case vchrHard1:
        if(bSimpleKeyb)
        {
          if(nRCSMode == 1) // Rot
           COMSendSimplex(COMMAND_RL);
         else if(nRCSMode == 2) // Lin
         {
           if(bInvDirect)
              COMSendSimplex(COMMAND_LB);
             else
              COMSendSimplex(COMMAND_LF);
         }
        }
        else
        {
         if(TimeGetMilliSeconds() > (ulTime + 300))// More that 300 ms elapsed
          {
            ulTime = TimeGetMilliSeconds();
            changeRCSmode();
          }
        }
       handled=true;
       break;
      case vchrHard2:
        if(bSimpleKeyb)
        {
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_Rl);
           else if(nRCSMode == 2) // Lin
             COMSendSimplex(COMMAND_LL);
        }
        else
        {
         if(nRCSMode == 1) // Rot
           COMSendSimplex(COMMAND_RL);
         else if(nRCSMode == 2) // Lin
         {
           if(bInvDirect)
              COMSendSimplex(COMMAND_LB);
             else
              COMSendSimplex(COMMAND_LF);
         }
        }
       handled=true;
       break;
      case vchrHard3:
        if(bSimpleKeyb)
        {
          if(nRCSMode == 1) // Rot
             COMSendSimplex(COMMAND_Rr);
           else if(nRCSMode == 2) // Lin
             COMSendSimplex(COMMAND_LR);
        }
        else
        {
         if(nRCSMode == 1) // Rot
           COMSendSimplex(COMMAND_RR);
         else if(nRCSMode == 2) // Lin
         {
           if(bInvDirect)
              COMSendSimplex(COMMAND_LF);
             else
              COMSendSimplex(COMMAND_LB);
         }
        }
       handled=true;
       break;
      case vchrHard4:
       if(bSimpleKeyb)
        {
         if(nRCSMode == 1) // Rot
           COMSendSimplex(COMMAND_RR);
         else if(nRCSMode == 2) // Lin
         {
           if(bInvDirect)
              COMSendSimplex(COMMAND_LF);
             else
              COMSendSimplex(COMMAND_LB);
         }
        }
       handled=true;
       break;
      default:
       break;
     }
    break;
   default:
    break;
  }

 if(bReceivingLed)
 {
   if((TimeGetMilliSeconds() - ulOldTime) > 500) // 500 ms elapsed
   {
     ulOldTime = TimeGetMilliSeconds();
     bReceivingLed = false;
     if(bRemoteOpen)
      DrawLed(BMPLEDON);
     else if(unPortId)
      DrawLed(BMPLEDLOCALON);
     else
      DrawLed(BMPLEDOFF);
   }
 }

 if(unPortId)
 {
   if(bKeepAlive)
     EvtResetAutoOffTimer();

   Errore = SrmReceiveCheck(unPortId, &lTotReadByte);
   if(lTotReadByte > 0)
   {
     lTotReadByte = SrmReceive(unPortId, cRxBuffer, RXBUFFERSIZE-1, 0, &Errore);
     if((Errore == errNone) || (Errore == serErrTimeOut))
     {
       if(lTotReadByte > 0)
       {
         DrawLed(BMPLEDRECEIVING);
         bReceivingLed = true;

         if(lTotReadByte >= RXBUFFERSIZE)
           lTotReadByte = RXBUFFERSIZE-1;

         cRxBuffer[lTotReadByte] = 0;

#ifdef _PRCoDEBUG
 if(hFile)
  FileWrite(hFile,cRxBuffer,lTotReadByte,1,&Errore);
#endif

         if((unBufferIx + lTotReadByte) < (BUFFERSIZE-1))
         {
           MemMove(&(cCommandBuffer[unBufferIx]), cRxBuffer, lTotReadByte);
           unBufferIx += lTotReadByte;
           cCommandBuffer[unBufferIx] = 0;
           bReinterpret = true;
         }
         else // Buffer full, we are getting rubbish
         {
           unBufferIx = 0;
           cCommandBuffer[0] = 0;
           bReinterpret = false;
         }

#ifdef _PRCoDEBUG
 if(hFile)
 {
  char cTemp[2];
  cTemp[0] = '{';
  FileWrite(hFile,cTemp,1,1,&Errore);
  FileWrite(hFile,cCommandBuffer,unBufferIx,1,&Errore);
  cTemp[0] = '}';
  FileWrite(hFile,cTemp,1,1,&Errore);
 }
#endif

         if(!bCanDraw)
           bReinterpret = false;

         while(bReinterpret)// more commands may be present in cCommandBuffer
         {
           bReinterpret = false;

           bCommandTerminator = false;
           unCommand = 0;
           for(unTempIx=0; unTempIx<unBufferIx; unTempIx++) // Search terminator
           {
             if(cCommandBuffer[unTempIx] == COMMANDTERMINATOR) // If command terminator found then exec cCommandBuffer
             {
               bCommandTerminator = true;
               cCommandBuffer[unTempIx] = 0; // terminete current command, other portions may exist after it
               unCommand = *(unsigned short *)cCommandBuffer; // interpreted as short for speed

#ifdef _PRCoDEBUG
 if(hFile)
 {
  char cTemp[2];
  cTemp[0] = '@';
  FileWrite(hFile,cTemp,1,1,&Errore);
 }
#endif
               break;
             }
           }

           if(bCommandTerminator)
           {
             bCommandTerminator = false;
             switch(unCommand)
              {
              case COMMAND_SV: // Vessel list
                emptyVesselList();
                unDestIx = 0;
                for(unIx=2; unIx<unTempIx; unIx++)
                {
                  if(cCommandBuffer[unIx] == STRINGTERMINATOR)
                  {
                    cVesselList[ulVesselListNum][unDestIx] = 0;
                    unDestIx = 0;
                    if(ulVesselListNum < (MAXNUMVESSELS-2))
                      ulVesselListNum++;
                  }
                  else
                  {
                    if(unDestIx < (MAXSIZEVESSELNAME-2))
                    {
                      cVesselList[ulVesselListNum][unDestIx] = cCommandBuffer[unIx];
                      unDestIx++;
                    }
                    else
                    {
                      cVesselList[ulVesselListNum][MAXSIZEVESSELNAME-1] = 0;
                    }
                  }
                }
                cVesselList[ulVesselListNum][MAXSIZEVESSELNAME-1] = 0;
                setFocusVessel();
                updateVesselList();
                break;
              case COMMAND_Ho: // Hover
                nSel = StrAToI((const char *)&(cCommandBuffer[2]));
                pScl = getObjectPtr(frmP,ScrollHover);
                SclSetScrollBar(pScl, 100-nSel, 0, 100, 10);
                ScrollBarHighlight(118,18,55,nSel,100,SBDOWNUP);
                break;
              case COMMAND_Ma: // Main
                nSel = StrAToI((const char *)&(cCommandBuffer[2]));
                pScl = getObjectPtr(frmP,ScrollMain);
                SclSetScrollBar(pScl, nSel, 0, 100, 10);
                ScrollBarHighlight(35,66,73,nSel,100,SBLEFTRIGHT);
                pScl = getObjectPtr(frmP,ScrollRetro);
                SclSetScrollBar(pScl, 100, 0, 100, 10);
                break;
              case COMMAND_Rt: // Retro
                nSel = StrAToI((const char *)&(cCommandBuffer[2]));
                pScl = getObjectPtr(frmP,ScrollRetro);
                SclSetScrollBar(pScl, 100-nSel, 0, 100, 10);
                ScrollBarHighlight(35,49,27,nSel,100,SBRIGHTLEFT);
                pScl = getObjectPtr(frmP,ScrollMain);
                SclSetScrollBar(pScl, 0, 0, 100, 10);
                break;
              case COMMAND_Se: // Set Focus to vessel
                MemMove(cFocusVessel, &(cCommandBuffer[2]), MAXSIZEVESSELNAME-1);
                cFocusVessel[MAXSIZEVESSELNAME-1] = 0;
                setFocusVessel();
                drawVesselList();
                break;
              case COMMAND_KR: // KillRot
                if(StrAToI((const char *)&(cCommandBuffer[2])) == 0)
                {
                  bCMDKillRot = false;
                  WinDrawChars("KillRot",7,16,131);
                }
                else
                {
                  bCMDKillRot = true;
                  WinDrawInvertedChars("KillRot",7,16,131);
                }
                break;
              case COMMAND_HL: // Set HLevel
                if(StrAToI((const char *)&(cCommandBuffer[2])) == 0)
                {
                  bCMDHLevel = false;
                  WinDrawChars("HLevel",6,4,146);
                }
                else
                {
                  bCMDHLevel = true;
                  WinDrawInvertedChars("HLevel",6,4,146);
                }
                break;
              case COMMAND_HA: // Set HAlt
                if(StrAToI((const char *)&(cCommandBuffer[2])) == 0)
                {
                  bCMDHAlt = false;
                  WinDrawChars("HAlt",4,38,146);
                }
                else
                {
                  bCMDHAlt = true;
                  WinDrawInvertedChars("HAlt",4,38,146);
                }
                break;
              case COMMAND_Pr: // Set Prograde
                if(StrAToI((const char *)&(cCommandBuffer[2])) == 0)
                {
                  bCMDPrograde = false;
                  WinDrawChars("Prograde",8,64,131);
                }
                else
                {
                  bCMDPrograde = true;
                  WinDrawInvertedChars("Prograde",8,64,131);
                }
                break;
              case COMMAND_Re: // Set Retrograde
                if(StrAToI((const char *)&(cCommandBuffer[2])) == 0)
                {
                  bCMDRetrograde = false;
                  WinDrawChars("Retrogr.",8,66,146);
                }
                else
                {
                  bCMDRetrograde = true;
                  WinDrawInvertedChars("Retrogr.",8,66,146);
                }
                break;
              case COMMAND_NP: // Set Normal +
                if(StrAToI((const char *)&(cCommandBuffer[2])) == 0)
                {
                  bCMDNormalPlus = false;
                  WinDrawChars("Normal (+)",10,111,131);
                }
                else
                {
                  bCMDNormalPlus = true;
                  WinDrawInvertedChars("Normal (+)",10,111,131);
                }
                break;
              case COMMAND_NM: // Set Normal -
                if(StrAToI((const char *)&(cCommandBuffer[2])) == 0)
                {
                  bCMDNormalMinus = false;
                  WinDrawChars("Normal (-)",10,112,146);
                }
                else
                {
                  bCMDNormalMinus = true;
                  WinDrawInvertedChars("Normal (-)",10,112,146);
                }
                break;
              case COMMAND_Le: // Set RCS Level
                nSel = StrAToI((const char *)&(cCommandBuffer[2]));
                pScl = getObjectPtr(frmP,ScrollRCSLevel);
                SclSetScrollBar(pScl, nSel, 0, 100, 10);
                ScrollBarHighlight(116,76,61,nSel,100,SBLEFTRIGHT);
                break;
              case COMMAND_SR: // Set RCS Mode
                nRCSMode = StrAToI((const char *)&(cCommandBuffer[2]));
                switch(nRCSMode)
                {
                case 0: // None
                  nRCSMode = 0;
                  WinDrawChars("Rot",3,59,49);
                  WinDrawChars("Lin",3,124,49);
                  break;
                case 1: // Rot
                  nRCSMode = 1;
                  WinDrawInvertedChars("Rot",3,59,49);
                  WinDrawChars("Lin",3,124,49);
                  break;
                case 2: // Lin
                  nRCSMode = 2;
                  WinDrawChars("Rot",3,59,49);
                  WinDrawInvertedChars("Lin",3,124,49);
                  break;
                }
                break;
              case COMMAND_CO: // Ready to COM
                bRemoteOpen = true;
                DrawLed(BMPLEDON);
                COMSendSimplex(COMMAND_GV); // Get Vessel list
                COMSendSimplex(COMMAND_GF); // Get Focuses vessel
                COMSendSimplex(COMMAND_GH); // Get Hover
                COMSendSimplex(COMMAND_GM); // Get Main
                COMSendSimplex(COMMAND_GR); // Get Retro
                COMSendSimplex(COMMAND_GL); // Get RCS Level
                COMSendSimplex(COMMAND_Gi); // Get KillRot
                COMSendSimplex(COMMAND_Ge); // Get HLevel
                COMSendSimplex(COMMAND_Gh); // Get HAlt
                COMSendSimplex(COMMAND_Go); // Get Prograde
                COMSendSimplex(COMMAND_Gt); // Get Retrograde
                COMSendSimplex(COMMAND_Gp); // Get Normal +
                COMSendSimplex(COMMAND_Gm); // Get Normal -
                COMSendSimplex(COMMAND_GT); // Get RCS type
                break;
              case COMMAND_CL:
                emptyVesselList();
                updateVesselList();
                bRemoteOpen = false;
                DrawLed(BMPLEDLOCALON);
                unBufferIx = 0;
                cCommandBuffer[0] = 0;
                break;
              }

              if(unBufferIx > (unTempIx+1))//Copy the residual command
              {
                MemMove(cCommandBuffer, &(cCommandBuffer[unTempIx+1]), unBufferIx-unTempIx-1);
                unBufferIx = unBufferIx-unTempIx-1;
                cCommandBuffer[unBufferIx] = 0;
                bReinterpret = true;
              }
              else
              {
                unBufferIx = 0;
                cCommandBuffer[0] = 0;
              }

#ifdef _PRCoDEBUG
 if(hFile)
 {
  char cTemp[2];
  cTemp[0] = '{';
  FileWrite(hFile,cTemp,1,1,&Errore);
  FileWrite(hFile,cCommandBuffer,unBufferIx,1,&Errore);
  cTemp[0] = '}';
  FileWrite(hFile,cTemp,1,1,&Errore);
 }
#endif

           } // if command terminator
         } // while
       } // if received data
     }
     else
     {
       CloseSerial();
       FrmCustomAlert(AlertGenericErr,"Error receiving from serial",NULL,NULL);
     }
   }
 }

 return handled;
}

/************************************/
void SetKeyToOff(UInt16 unType)
{
  if(!bCanDraw)
    return;

  if(unType != COMMAND_KR)
  {
    bCMDKillRot = false;
    WinDrawChars("KillRot",7,16,131);
  }
  if(unType != COMMAND_Pr)
  {
    bCMDPrograde = false;
    WinDrawChars("Prograde",8,64,131);
  }
  if(unType != COMMAND_Re)
  {
    bCMDRetrograde = false;
    WinDrawChars("Retrogr.",8,66,146);
  }
  if(unType != COMMAND_HL)
  {
    bCMDHLevel = false;
    WinDrawChars("HLevel",6,4,146);
  }
  if(unType != COMMAND_NP)
  {
    bCMDNormalPlus = false;
    WinDrawChars("Normal (+)",10,111,131);
  }
  if(unType != COMMAND_NM)
  {
    bCMDNormalMinus = false;
    WinDrawChars("Normal (-)",10,112,146);
  }
  if(unType != COMMAND_HA)
  {
    bCMDHAlt = false;
    WinDrawChars("HAlt",4,38,146);
  }
}

/**************************/
/* Draw Vessels list item */
/**************************/
void ListDraw(Int16 itemNum,RectangleType *bounds,Char **itemsText)
{
 if(itemsText)
  if(itemsText[itemNum])
    if(bCanDraw)
      WinDrawTruncChars(itemsText[itemNum],StrLen(itemsText[itemNum]),
	                       bounds->topLeft.x,bounds->topLeft.y,bounds->extent.x);
}

/********************/
/* Set Vessels list */
/********************/
void	setList(Char **itemsText,Int16 numItems,UInt16 unID,UInt16 unDim)
{
 FormPtr pForm=FrmGetActiveForm();
 void	*pList=getObjectPtr(pForm,unID);

 LstSetListChoices(pList,itemsText,numItems);
 LstSetHeight(pList,numItems<unDim ? numItems : unDim);
 LstSetDrawFunction(pList,(ListDrawDataFuncPtr)ListDraw);
}

/**********************/
/* Draw Vessels list  */
/**********************/
void drawVesselList()
{
  FormPtr   frmP = FrmGetActiveForm();
  void	     *pList;

  if(unPortId)
  {
    pList=getObjectPtr(frmP,VesselList);
    LstSetSelection(pList,unVesselListItem);
    if(bCanDraw)
    {
      WinDrawChars(cSpaces,44,19,2);
      WinDrawTruncChars(ppVesselList[unVesselListItem],StrLen(ppVesselList[unVesselListItem]),19,2,90);
    }
  }
}

/***********************/
/* Update Vessel list  */
/***********************/
void updateVesselList()
{
  setList(ppVesselList,(ulVesselListNum<14)?ulVesselListNum:14,VesselList,10);
  drawVesselList();
}

/**************************************************/
void GetListItem(UInt16 unListID, char * cStr)
{
  UInt16  unI;
  ListPtr  pList;
  char    *cTemp;
  FormPtr frmP=FrmGetActiveForm();

  pList=getObjectPtr(frmP,unListID);
  unI=LstGetSelection(pList);
  if(unI<ulVesselListNum)
   {
    cTemp=LstGetSelectionText(pList,unI);
    if(cTemp)
     StrCopy(cStr, cTemp);
   }
}

/************************************/
/* MainForm events handled here     */
/************************************/
Boolean mainFormEventHandler(EventPtr eventP)
{
 Boolean handled = false;
 FormPtr frmP = FrmGetActiveForm();
 UInt16 unFrmID;
 Int16  nSel;
 ScrollBarPtr     pScl;

 switch(eventP->eType)
  {
   case frmOpenEvent:// open a form
    FrmDrawForm(frmP);
    unFrmID=FrmGetActiveFormID();
    if(unFrmID==MainForm)// it's the main form
     {
        newRgb.r=0x00;
        newRgb.g=0x00;
        newRgb.b=0xFF;
        WinSetForeColorRGB(&newRgb,&prevRgb);
        Rect.topLeft.x=2;
        Rect.topLeft.y=17;
        Rect.extent.x=156;
        Rect.extent.y=28;
        WinDrawRectangleFrame(rectangleFrame,&Rect);
        Rect.topLeft.x=2;
        Rect.topLeft.y=46;
        Rect.extent.x=35;
        Rect.extent.y=82;
        WinDrawRectangleFrame(rectangleFrame,&Rect);
        Rect.topLeft.x=38;
        Rect.topLeft.y=46;
        Rect.extent.x=120;
        Rect.extent.y=82;
        WinDrawRectangleFrame(rectangleFrame,&Rect);
        newRgb.r=0xFF;
        newRgb.g=0xFF;
        newRgb.b=0xFF;
        WinSetForeColorRGB(&newRgb,&prevRgb);
        WinDrawLine(2, 45, 36, 45);
        if(unPortId)
        {
          DrawLed(BMPLEDLOCALON);
          WinDrawChars(cSpaces,44,19,2);
          WinDrawChars(cFocusVessel,StrLen(cFocusVessel),19,2);
        }
        else
        {
          DrawLed(BMPLEDOFF);
          WinDrawChars("< COM Inactive >",16,19,2);
        }

       if(nRCSMode == 1) // Rot
       {
          WinDrawInvertedChars("Rot",3,59,49);
          WinDrawChars("Lin",3,124,49);
       }
       else if(nRCSMode == 2) // Lin
       {
          WinDrawChars("Rot",3,59,49);
          WinDrawInvertedChars("Lin",3,124,49);
       }
       else
       {
          WinDrawChars("Rot",3,59,49);
          WinDrawChars("Lin",3,124,49);
       }
       ScrollBarHighlight(116,76,61,100,100,SBLEFTRIGHT);
     }
    else if(unFrmID==PrefsForm)
    {
      SetValue32(PrefsPort, prefs.ulSerialPort);
      SetValue32(PrefsBaudRate, prefs.ulBaudRate);
      FrmSetControlValue(frmP,FrmGetObjectIndex(frmP,PrefsKeepAlive),bKeepAlive);
      FrmSetControlValue(frmP,FrmGetObjectIndex(frmP,PrefsInvDirect),bInvDirect);
      bNewSimpleKeyb = bSimpleKeyb;
      if(bSimpleKeyb)
        FrmSetControlGroupSelection(frmP,1,PrefsSimple);
      else
        FrmSetControlGroupSelection(frmP,1,Prefs5Way);
    }
    handled = true; 
    break; 
   case menuEvent: 
    handled = doMenu(frmP, eventP->data.menu.itemID); 
    break; 
   case ctlSelectEvent:
     switch (eventP->data.ctlSelect.controlID)
     {
     case Prefs5Way:
       bNewSimpleKeyb = false;
       handled = true; 
       break;
     case PrefsSimple:
       bNewSimpleKeyb = true;
       handled = true; 
       break;
     case PrefsOk:
       prefs.ulSerialPort = GetValue32(PrefsPort);
       switch(prefs.ulSerialPort)
       {
       case 0: prefs.ulSerialPort = 32768; break;//  Physical HotSync (RS-232 or USB).
       case 1: prefs.ulSerialPort = 32769; break;//  IR port.
       case 2: prefs.ulSerialPort = 32770; break;//  Debug console (RS-232 or USB).
       case 3: prefs.ulSerialPort = 32771; break;//  RS-232 cradle.
       case 4: prefs.ulSerialPort = 32772; break;//  USB cradle.
       case 5: prefs.ulSerialPort = 1966289464; break;//  Cradle port using the 68328 UART.
       case 6: prefs.ulSerialPort = 1966622042; break;//  Cradle port using the 68328EZ UART.
       case 7: prefs.ulSerialPort = 1966486832; break;//  Specifies the IR port on the upgrade card.
       case 8: prefs.ulSerialPort = 1651794800; break;//  Bluetooth Connection virtual.
       case 9: prefs.ulSerialPort = 1769104237; break;//  IrDA IRComm protocol virtual.
       case 10: prefs.ulSerialPort = 1919312749; break;//  Bluetooth RFCOMM virtual.
       }

       prefs.ulBaudRate = GetValue32(PrefsBaudRate);

       bKeepAlive = FrmGetControlValue(frmP,FrmGetObjectIndex(frmP,PrefsKeepAlive));
       if(bKeepAlive)
        prefs.ulControlBits |= CBITS_KEEPALIVE;
       else
        prefs.ulControlBits &= ~CBITS_KEEPALIVE;

       bInvDirect = FrmGetControlValue(frmP,FrmGetObjectIndex(frmP,PrefsInvDirect));
       if(bInvDirect)
        prefs.ulControlBits |= CBITS_INVDIRECT;
       else
        prefs.ulControlBits &= ~CBITS_INVDIRECT;

       bSimpleKeyb = bNewSimpleKeyb;
       if(bSimpleKeyb)
        prefs.ulControlBits |= CBITS_SIMPLEKEYB;
       else
        prefs.ulControlBits &= ~CBITS_SIMPLEKEYB;

       PrefSetAppPreferences(AppCreator,0,1,&prefs,sizeof(Prefs),true);
       FrmGotoForm(MainForm);
       handled = true; 
       break;
     case AboutFormOK:
     case PrefsClear:
       FrmGotoForm(MainForm);
       handled = true; 
       break;
     case CMDSetFocus:
       COMSendSimplex(COMMAND_SF);
       handled = true; 
       break;
     case CMDRCSLin: // RCS mode Lin
       nRCSMode = 2;
       WinDrawChars("Rot",3,59,49);
       WinDrawInvertedChars("Lin",3,124,49);
       COMSendNumber(COMMAND_SR, nRCSMode);
       handled = true; 
       break;
     case CMDRCSRot: // RCS mode Rot
       nRCSMode = 1;
       WinDrawInvertedChars("Rot",3,59,49);
       WinDrawChars("Lin",3,124,49);
       COMSendNumber(COMMAND_SR, nRCSMode);
       handled = true; 
       break;
     case CMDRCSOff: // RCS mode Off
       nRCSMode = 0;
       WinDrawChars("Rot",3,59,49);
       WinDrawChars("Lin",3,124,49);
       COMSendNumber(COMMAND_SR, nRCSMode);
       handled = true; 
       break;
     case CMDKillRot:
       if(bCMDKillRot)
       {
         bCMDKillRot = false;
         COMSendNumber(COMMAND_KR,0);
         SetKeyToOff(COMMAND_KR);
         COMSendSimplex(COMMAND_Ge); // Get HLevel
         COMSendSimplex(COMMAND_Gh); // Get HAlt
         WinDrawChars("KillRot",7,16,131);
       }
       else
       {
         bCMDKillRot = true;
         COMSendNumber(COMMAND_KR,1);
         SetKeyToOff(COMMAND_KR);
         COMSendSimplex(COMMAND_Ge); // Get HLevel
         COMSendSimplex(COMMAND_Gh); // Get HAlt
         WinDrawInvertedChars("KillRot",7,16,131);
       }
       handled = true; 
       break;
     case CMDPrograde:
       if(bCMDPrograde)
       {
         bCMDPrograde = false;
         COMSendNumber(COMMAND_Pr,0);
         SetKeyToOff(COMMAND_Pr);
         WinDrawChars("Prograde",8,64,131);
       }
       else
       {
         bCMDPrograde = true;
         COMSendNumber(COMMAND_Pr,1);
         SetKeyToOff(COMMAND_Pr);
         WinDrawInvertedChars("Prograde",8,64,131);
       }
       handled = true; 
       break;
     case CMDRetrograde:
       if(bCMDRetrograde)
       {
         bCMDRetrograde = false;
         COMSendNumber(COMMAND_Re,0);
         SetKeyToOff(COMMAND_Re);
         WinDrawChars("Retrogr.",8,66,146);
       }
       else
       {
         bCMDRetrograde = true;
         COMSendNumber(COMMAND_Re,1);
         SetKeyToOff(COMMAND_Re);
         WinDrawInvertedChars("Retrogr.",8,66,146);
       }
       handled = true; 
       break;
     case CMDHLevel:
       if(bCMDHLevel)
       {
         bCMDHLevel = false;
         COMSendNumber(COMMAND_HL,0);
         SetKeyToOff(COMMAND_HL);
         COMSendSimplex(COMMAND_Gi); // Get KillRot
         COMSendSimplex(COMMAND_Gh); // Get HAlt
         WinDrawChars("HLevel",6,4,146);
       }
       else
       {
         bCMDHLevel = true;
         COMSendNumber(COMMAND_HL,1);
         SetKeyToOff(COMMAND_HL);
         COMSendSimplex(COMMAND_Gi); // Get KillRot
         COMSendSimplex(COMMAND_Gh); // Get HAlt
         WinDrawInvertedChars("HLevel",6,4,146);
       }
       handled = true; 
       break;
     case CMDNormalPlus:
       if(bCMDNormalPlus)
       {
         bCMDNormalPlus = false;
         COMSendNumber(COMMAND_NP,0);
         SetKeyToOff(COMMAND_NP);
         WinDrawChars("Normal (+)",10,111,131);
       }
       else
       {
         bCMDNormalPlus = true;
         COMSendNumber(COMMAND_NP,1);
         SetKeyToOff(COMMAND_NP);
         WinDrawInvertedChars("Normal (+)",10,111,131);
       }
       handled = true; 
       break;
     case CMDNormalMinus:
       if(bCMDNormalMinus)
       {
         bCMDNormalMinus = false;
         COMSendNumber(COMMAND_NM,0);
         SetKeyToOff(COMMAND_NM);
         WinDrawChars("Normal (-)",10,112,146);
       }
       else
       {
         bCMDNormalMinus = true;
         COMSendNumber(COMMAND_NM,1);
         SetKeyToOff(COMMAND_NM);
         WinDrawInvertedChars("Normal (-)",10,112,146);
       }
       handled = true; 
       break;
     case CMDHAlt:
       if(bCMDHAlt)
       {
         bCMDHAlt = false;
         COMSendNumber(COMMAND_HA,0);
         SetKeyToOff(COMMAND_HA);
         COMSendSimplex(COMMAND_Gi); // Get KillRot
         COMSendSimplex(COMMAND_Ge); // Get HLevel
         WinDrawChars("HAlt",4,38,146);
       }
       else
       {
         bCMDHAlt = true;
         COMSendNumber(COMMAND_HA,1);
         SetKeyToOff(COMMAND_HA);
         COMSendSimplex(COMMAND_Gi); // Get KillRot
         COMSendSimplex(COMMAND_Ge); // Get HLevel
         WinDrawInvertedChars("HAlt",4,38,146);
       }
       handled = true; 
       break;
      case CMDShutdown: // Shut Main, Retro and Hover engines
       COMSendNumber(COMMAND_Ma, 0);
       COMSendNumber(COMMAND_Rt, 0);
       COMSendNumber(COMMAND_Ho, 0);
       pScl = getObjectPtr(frmP,ScrollHover);
       SclSetScrollBar(pScl, 100, 0, 100, 10);
       pScl=getObjectPtr(frmP,ScrollMain);
       SclSetScrollBar(pScl, 0, 0, 100, 10);
       pScl=getObjectPtr(frmP,ScrollRetro);
       SclSetScrollBar(pScl, 100, 0, 100, 10);
       break;
     default:
       break;
     }
    break;
   case popSelectEvent:
      if (eventP->data.popSelect.controlID == VesselPopTrig)
       {
        unVesselListItem=eventP->data.popSelect.selection;
        drawVesselList();
        GetListItem(VesselList, cFocusVessel);
        COMSendString(COMMAND_Se, cFocusVessel);
       }
     handled = true;
     break;
   case sclRepeatEvent:
    nSel=eventP->data.sclRepeat.newValue;
    switch(eventP->data.sclRepeat.scrollBarID)
    {
      case ScrollHover:
        COMSendNumber(COMMAND_Ho, 100-nSel);
        ScrollBarHighlight(118,18,55,100-nSel,100,SBDOWNUP);
        break;
      case ScrollRetro:
        COMSendNumber(COMMAND_Ma, 0);
        COMSendNumber(COMMAND_Rt, 100-nSel);
        ScrollBarHighlight(35,49,27,100-nSel,100,SBRIGHTLEFT);
        pScl=getObjectPtr(frmP,ScrollMain);
        SclSetScrollBar(pScl, 0, 0, 100, 10);
        break;
      case ScrollMain:
        COMSendNumber(COMMAND_Rt, 0);
        COMSendNumber(COMMAND_Ma, nSel);
        ScrollBarHighlight(35,66,73,nSel,100,SBLEFTRIGHT);
        pScl=getObjectPtr(frmP,ScrollRetro);
        SclSetScrollBar(pScl, 100, 0, 100, 10);
        break;
      case ScrollRCSLevel:
        COMSendNumber(COMMAND_Le, nSel);
        ScrollBarHighlight(116,76,61,nSel,100,SBLEFTRIGHT);
        break;
    }
    break;
   case ctlRepeatEvent:
     switch(eventP->data.ctlRepeat.controlID)
     {
     case CMDLinX: //  RCS Lin KillRot
     case CMDRotX: //  RCS Rot KillRot
       bCMDKillRot = true;
       WinDrawInvertedChars("KillRot",7,16,131);
       COMSendSimplex(COMMAND_RX);
       COMSendSimplex(COMMAND_Ge); // Get HLevel
       COMSendSimplex(COMMAND_Gh); // Get HAlt
       break;
     case CMDRotD: //  RCS Rot Down
       COMSendSimplex(COMMAND_RD);
       break;
     case CMDRotR: //  RCS Rot Right
       COMSendSimplex(COMMAND_RR);
       break;
     case CMDRotU: //  RCS Rot Up
       COMSendSimplex(COMMAND_RU);
       break;
     case CMDRotL: //  RCS Rot Left
       COMSendSimplex(COMMAND_RL);
       break;
     case CMDRotRr: //  RCS Rot Right ^
       COMSendSimplex(COMMAND_Rr);
       break;
     case CMDRotRl: // RCS Rot Left ^
       COMSendSimplex(COMMAND_Rl);
       break;
     case CMDLinF: // RCS Lin 
       if(bInvDirect)
         COMSendSimplex(COMMAND_LF);
       else
         COMSendSimplex(COMMAND_LB);
       break;
     case CMDLinB: // RCS Lin 
       if(bInvDirect)
         COMSendSimplex(COMMAND_LB);
       else
         COMSendSimplex(COMMAND_LF);
       break;
     case CMDLinD: // RCS Lin 
       if(bInvDirect)
        COMSendSimplex(COMMAND_LU);
       else
        COMSendSimplex(COMMAND_LD);
       break;
     case CMDLinR: // RCS Lin 
       COMSendSimplex(COMMAND_LR);
       break;
     case CMDLinU: // RCS Lin 
       if(bInvDirect)
        COMSendSimplex(COMMAND_LD);
       else
        COMSendSimplex(COMMAND_LU);
       break;
     case CMDLinL: // RCS Lin 
       COMSendSimplex(COMMAND_LL);
       break;
     default:
       break;
     }
     break;
   default:
     break;
  }
 return handled;
}

/************************************/
/* Application events handled here  */
/************************************/
Boolean appHandleEvent(EventPtr event)
{
 FormPtr  frm;
 Int16    formId;
 Boolean  handled = false;

 if(event->eType == frmLoadEvent) /* Form loading */
  { 
   formId = event->data.frmLoad.formID; 
   frm = FrmInitForm(formId); 
   FrmSetActiveForm(frm); 

   if((formId==MainForm) || (formId==AboutForm) || (formId==PrefsForm))
    FrmSetEventHandler(frm,mainFormEventHandler);

   handled = true;
  }
 return handled;
}

/************************************/
/* Called at start                  */
/************************************/
void startApp()
{
  UInt16 prefSize = sizeof(Prefs);

  unPortId = 0;
  Errore = 0;
  bCMDHLevel = false;
  bCMDKillRot = false;
  bCMDPrograde = false;
  bCMDRetrograde = false;
  bCMDNormalPlus = false;
  bCMDNormalMinus = false;
  bCMDHAlt = false;
  nRCSMode = 1; // Rot
  unBufferIx = 0;
  bKeepAlive = false;
  bSimpleKeyb = false;
  bInvDirect = false;
  bReceivingLed = false;
  bRemoteOpen = false;
  bCanDraw = true;
  MemSet(cSpaces,sizeof(cSpaces),' ');
  cSpaces[sizeof(cSpaces)-1]=0;
  cFocusVessel[0] = 0;
  for(unVesselListItem=0; unVesselListItem<MAXNUMVESSELS; unVesselListItem++)
  {
   ppVesselList[unVesselListItem]=(Char *)&(cVesselList[unVesselListItem]);
   cVesselList[unVesselListItem][0]=0;
  }
  ulVesselListNum = 0;
  unVesselListItem = 0;

  //Load application preferences
  if((PrefGetAppPreferences(AppCreator,0,&prefs,&prefSize,true)==noPreferenceFound)
    ||(prefSize != sizeof(Prefs)))
  {
    prefs.ulSerialPort  = 1919312749; // Default to Bluetooth
    prefs.ulBaudRate    = 115200;
    prefs.ulControlBits = 0;
    prefs.unKeybMode    = 0;

    PrefSetAppPreferences(AppCreator,0,1,&prefs,sizeof(Prefs),true);
  }
  else
  {
    if(prefs.ulControlBits & CBITS_KEEPALIVE)
      bKeepAlive = true;
    else
      bKeepAlive = false;

    if(prefs.ulControlBits & CBITS_INVDIRECT)
      bInvDirect = true;
    else
      bInvDirect = false;

    if(prefs.ulControlBits & CBITS_SIMPLEKEYB)
      bSimpleKeyb = true;
    else
      bSimpleKeyb = false;
  }

  unTickPerSec = SysTicksPerSecond();//Comes in handy to compute a more resolute time
  ulTime = TimeGetMilliSeconds();
  ulOldTime = ulTime;

  hBuffer=MemHandleNew(BUFFERSIZE+1);
  if(hBuffer)
  {
    cCommandBuffer=MemHandleLock(hBuffer);
    cCommandBuffer[0]=0;
  }
  else
  {
    cCommandBuffer = NULL;
  }

  KeyRates(false, &initDelayDef, &periodDef, &doubleTapDelayDef, &queueAheadDef);
  bKeyRatesUp = false;
}

/************************************/
/* Called at exit                   */
/************************************/
void stopApp()
{
  KeyRates(true, &initDelayDef, &periodDef, &doubleTapDelayDef, &queueAheadDef);
  CloseSerial();

  if(hBuffer)
  {
    MemHandleUnlock(hBuffer);
    MemHandleFree(hBuffer);
  }
}

/************************************/
/* The main program function        */
/************************************/
UInt32 PilotMain(UInt16 cmd,void *cmdPBP,UInt16 launchFlags)
{
 EventType event;
 UInt16    error;

 switch(cmd)
  {
   case sysAppLaunchCmdNormalLaunch:
	   error=RomVersionCompatible(MinVersion,launchFlags);
    if(error==0)
     {
      startApp();
      FrmGotoForm(MainForm);
      do 
       { 
        EvtGetEvent(&event,20); 

        if(PreprocessEvent(&event)) 
         continue; 

        if(!SysHandleEvent(&event)) 
         if(!MenuHandleEvent(0,&event,&error)) 
          if(!appHandleEvent(&event)) 
           FrmDispatchEvent(&event); 
       } 
      while(event.eType != appStopEvent);
      bCanDraw = false;
      FrmCloseAllForms();
      stopApp();
     }
    break;
  }
 return 0; 
}
